#!/usr/bin/env python

"""Tests for `configuretron` package."""

import pytest

from click.testing import CliRunner

from configuretron import configuretron
from configuretron import cli
from dataclasses import dataclass
import os.path


@dataclass
class BasicConfig:
    some_int: int
    some_float: float
    some_dict: dict
    some_list_numbers: list[int]
    some_list_strings: list[str]
    test_inherit: str
    test_override: str


@dataclass
class EncryptedConfig:
    secret: str
    secret_list: list[int]


private_key = b"-----BEGIN RSA PRIVATE KEY-----\nMIIEqwIBAAKCAQEAh7OC5Ki6VssdeRoGiQSKQs9BULEAQXHCoq7lnfmgzQ6gSlSw\nd9vBxU0zrv8eGZRGzKFM0C8i7RU4kEf8QhYrfMu8noclBlqQzPOy8uKBHOsL6Z5v\nCAt01KP8REC5UG4Aj3zEADsO4Ne/vTWkdNQi8y/zOt5lC7TJ3Hkpa/3KvVVNKZcZ\nJcyQcDRc2IKbaSnlN2WdyjkzgZ6LQ6mLl961++kBDdmFXgyaeREFJpIsq+QUGYZz\nvsK8bUqL2tCbeaX/ZaDhgjEnjk6KzO66tvgBT8OuYjbukiwdeyfCMamAuB4BMLO1\nmCWi5u99K5xfVA176abjWxyYup5JWMKNbOurgwIDAQABAoIBAAwBX6UnIEUpdxQN\nr2JT2n+KonanEiS4EcYevwW43fcGQjwRPgN8l6oC6H2k6F7O9WyJIKiSYv6ij/yD\nwdd+6p4B0FAeGLJ1NLZNIRnH3DBYwxcKAzys9ssRLvJC36Iz+Sy4lwaGMwzFSZtN\nsZz8X5MwPM6KeloMxnYn8xKJjAT2pGjhBnI6g+G+XeWEl9AW55FFLvPIypRv9Lcz\nURhHhR5wJMaRuReSwSqJLO1CUVd71mrxIMGhFLhDAA3/r4fPuywYlmzhGINLhDni\nWWIioM6OMJosskpv9cmyi/EzhBhCnfC5C7PtLLqmtWneEkJEsxhAZXHAnKk+Oc2R\n66A35oECgYkAvyMvflTs+tp+uTOEvNlGgv/RuSZQkVU7VFQOTxWkOseZD5bdRDYc\niKtX7w855zSxr+zEVI/H8dnfh0LKTMSU3joS/4TiwWe+bNYQqLzYimV7sg9gJjhK\nbiGXcPgt4QYNUhjCMFxwTntU3FewtJs+MZskYXRm+NbGPYRQQ5ZGGfvkG+GB6IiJ\n4wJ5ALXAYFUUrmIlPkGAP+WNdNLVRrREuvMoEROpWo0zpxC+9bihrFX+NjwaT1ki\nNwLoXEzyrNuuf5VHx3CLMwSgxQj1by36jhlkWdzDH6MMUggeCiL5INmtyMxYSoOP\nwHG6G7OJxjSBChsvq3N6AZ9LkcJ+JjJSgmCJ4QKBiQCYZdeSZOhmwxvTUu59HMiJ\n3qs7cv8+QbUCcdrO17SQYWD6+xEFOBfcKLH+HCfQ30TlvmR0AAguH8eIM4rVVtBT\nt/452ZxrFOrSIIQ75gbJokzUFkpVbwB5ezMikd8S0h6A3NjIcovhJ5jr7scn2bfn\nGYoSYhtQQP7jQcI1gXX3dc3VJSLNY0B9AnkAsAv/Q2oc34QjoV6QhdFW70EWk1Zf\nn1eX8ut/gnYdxOKzMHupakIqVl1FrpitoGthvEbzmHaVrgsw65ppeHohYGQbpPWk\n3oONZ4C5DD3K14IBX47gevkSHp0G0BhV19LlMiqpigHTwDGUaO4s1BdmwHVgmi3B\ni7EBAoGJAJItwE1Qi2fyIUNb6UQEbOrbXAN695JlojW7HvZ7cmQ0KYMD99/c2KIj\nhxA9CpZE/ukE4I203fasX9J2eb/oESfabC46Pau2k1pnW8t9XZx58qqX1pDlq/mT\nGJyyOZPuVVzaAabsjRfoYEHQM4Jm+IyH7h/c/ZcdYCTanLKuia1eDHCD6b+9LSA=\n-----END RSA PRIVATE KEY-----\n"

test_dir = os.path.dirname(os.path.realpath(__file__))
BASIC_CONFIG_YAML = os.path.join(test_dir, "config_basic.yml")
ENCRYPTED_CONFIG_YAML = os.path.join(test_dir, "config_encrypted.yml")


def test_basic_configuration():
    config = configuretron.from_yaml(BasicConfig, BASIC_CONFIG_YAML)
    assert config.some_int == 12345
    assert config.some_float == 1.11
    assert config.some_dict == {"key": "value", "other": "out"}
    assert config.some_list_numbers == [1, 2, 3]
    assert config.some_list_strings == ["wow", "it's a", "list"]
    assert config.test_inherit == "inherited"
    assert config.test_override == "global"


def test_environment_override():
    basic_config = configuretron.from_yaml(BasicConfig, BASIC_CONFIG_YAML)
    prod_config = configuretron.from_yaml(BasicConfig, BASIC_CONFIG_YAML, "prod")
    assert basic_config.test_inherit == "inherited"
    assert basic_config.test_override == "global"
    assert prod_config.test_inherit == "inherited"
    assert prod_config.test_override == "prod"


def test_encryption():
    config = configuretron.from_yaml(EncryptedConfig, ENCRYPTED_CONFIG_YAML, private_key=private_key)
    assert config.secret == "secret"
    assert config.secret_list == [1, 2, 3]


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.cli)
    assert result.exit_code == 0
    assert "Usage: " in result.output
    help_result = runner.invoke(cli.cli, ["--help"])
    assert help_result.exit_code == 0



"""Console script for configuretron."""
from base64 import b64encode, b64decode
import click
import json
import rsa
import sys
import yaml
from . import configuretron


def write_yaml(config_dict, yaml_path):
    out_yaml = yaml.dump(config_dict)
    click.echo(f"Saving YAML file {yaml_path}...", nl=False)
    try:
        open(yaml_path, "w").write(out_yaml)
    except Exception:
        click.echo("error writing to file")
        exit()
    click.echo("success")


@click.group()
@click.option("--yaml", "yaml_path", default=None, help="YAML configuration location")
@click.option("--private-key", default=None, help="Private key string")
@click.option("--private-key-file", default=None, help="Private key file path")
@click.option("--private-key-b64", default=None, help="Private key as a base64 string")
@click.option("--env", default=None, help="Environment to use")
@click.pass_context
def cli(ctx, yaml_path, private_key, private_key_file, private_key_b64, env):
    # -------- YAML --------
    if yaml_path is None and not "decrypt" in sys.argv:
        click.echo(f"Please provide a YAML configuration file location (--yaml=PATH before the provided command)")
        exit()

    ctx.ensure_object(dict)
    ctx.obj["yaml_path"] = yaml_path
    ctx.obj["env"] = env

    if yaml_path:
        click.echo(f"Loading YAML file {yaml_path}...", nl=False)
        try:
            ctx.obj["config"] = configuretron.yaml_to_dict(yaml_path)
        except Exception as e:
            click.echo(f"{e}")
            exit()

        click.echo(f"success")

    # -------- private key --------

    if private_key is not None:
        click.echo(f"Loading private key...", nl=False)
        ctx.obj["private_key"] = private_key
        click.echo(f"success")
    elif private_key_file is not None:
        click.echo(f"Reading private key file {private_key_file}...", nl=False)
        try:
            ctx.obj["private_key"] = open(private_key_file, "rb").read()
        except Exception as e:
            click.echo(f"Could not read private key file: {e}")
            exit()
        click.echo(f"success")
    elif private_key_b64 is not None:
        click.echo(f"Loading base64 private key...", nl=False)
        try:
            ctx.obj["private_key"] = b64decode(private_key_b64)
        except Exception as e:
            click.echo(f"Could not decode private key: {e}")
            exit()
        click.echo(f"success")
    else:
        ctx.obj["private_key"] = None


@cli.command()
@click.option(
    "--bits",
    type=int,
    default=2048,
    help="Number of bits used to generate the encryption key.  Higher numbers will result in higher generation times.",
)
@click.pass_context
def setup_encryption(ctx, bits):
    config_dict = ctx.obj["config"]
    yaml_path = ctx.obj["yaml_path"]

    # Validate key doesn't exist
    if "encryption" in config_dict and "key" in config_dict["encryption"]:
        click.echo("This configuration already includes an encryption key.")
        click.echo("Please remove encryption.key if you intend to overwrite it.")
        exit()

    if not "encyption" in config_dict:
        config_dict["encryption"] = {}

    # Generate pair
    click.echo(f"Generating {bits} bit RSA key pair...", nl=False)
    (key_public, key_private) = rsa.newkeys(bits, poolsize=8)
    config_dict["encryption"]["key"] = b64encode(key_public.save_pkcs1()).decode()
    click.echo("success")

    # Write to YAML
    write_yaml(config_dict, yaml_path)

    # Output private key
    private_key_pem = key_private.save_pkcs1()
    click.echo(f"\n===================")
    click.echo(f"=== Private Key ===")
    click.echo(f"===================\n")
    click.echo(f"Keep this secret and safe!  Without it you will not be able to decrypt any secrets:\n")
    click.echo(f"Code:")
    click.echo(f'   private_key = open("private_key.pem", "rb").read()\n')
    click.echo(f"Keyfile:\n")
    click.echo(private_key_pem)
    click.echo(f"\n============================")
    click.echo(f"=== Private Key B64 ENV  ===")
    click.echo(f"============================\n")
    click.echo(f"If passed as an environment variable (such as CONFIG_PRIVATE_KEY), use this base64 version instead\n")
    click.echo(f"Code:")
    click.echo(f'   private_key = configuretron.env_base64_value("CONFIG_PRIVATE_KEY")\n')
    click.echo(f"Key:\n")
    click.echo(b64encode(private_key_pem))


@cli.command()
@click.option("--key", default=None, help="Variable name to encrypt")
@click.argument("values", nargs=-1)
@click.pass_context
def encrypt(ctx, key=None, values=None):
    config_dict = ctx.obj["config"]
    config_args = config_dict["config"]
    yaml_path = ctx.obj["yaml_path"]

    # Validate key exists
    if "encryption" not in config_dict or "key" not in config_dict["encryption"]:
        click.echo("Encryption has not been set up for this configuration.  Please run setup_encryption.")
        exit()

    try:
        public_key_pem = b64decode(config_dict["encryption"]["key"])
        public_key = rsa.PublicKey.load_pkcs1(public_key_pem)
    except Exception:
        click.echo("Unable to process public key from config")

    if key:
        if len(values) > 1:
            value = values
        elif values:
            value = values[0]
        else:
            if not key in config_args:
                click.echo(f"Config is missing key {key}")
                exit()
            value = config_args[key]

        click.echo(f"Encrypting {key} as {value}...", nl=False)
        config_args[f"{key}.encrypted"] = configuretron.encrypt(value, public_key)
        click.echo(f"success")

        # Clear old key
        if key in config_args:
            del config_args[key]

        # Write to YAML
        write_yaml(config_dict, yaml_path)
    else:
        for value in values:
            click.echo(f"\n====================")
            click.echo(f"{value}")
            click.echo(f"====================")
            click.echo(f"{configuretron.encrypt(value, public_key)}")
            click.echo(f"====================")


@cli.command()
@click.argument("value")
@click.pass_context
def decrypt(ctx, value):
    if ctx.obj["private_key"] is None:
        click.echo(f"Please provide a private key (before the provided command)")
        exit()

    try:
        rsa_private_key = rsa.PrivateKey.load_pkcs1(ctx.obj["private_key"])
    except Exception as e:
        click.echo(f"Could not load private key: {e}")
        exit()

    try:
        value_decrypted = configuretron.decrypt(value, rsa_private_key)
    except Exception as e:
        click.echo(f"Error: {e}")
        exit()

    click.echo("Value:\n")
    click.echo(value_decrypted)


@cli.command()
@click.pass_context
def dump(ctx):
    try:
        config_args = configuretron.dict_to_args(
            config_dict=ctx.obj["config"], env=ctx.obj["env"], private_key=ctx.obj["private_key"]
        )
    except Exception as e:
        click.echo(f"Error: {e}")
        exit()
    click.echo(json.dumps(config_args, indent=4))


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



from base64 import b64decode, b64encode
import json
from os import environ, path
import rsa
from typing import Any, TypeVar
import yaml

T = TypeVar("T")


def from_yaml(config_class: T, file: str, env: str = None, private_key: bytes = None) -> T:
    config_dict = yaml_to_dict(file)

    return from_dict(config_class=config_class, config_dict=config_dict, env=env, private_key=private_key)


def from_dict(config_class: T, config_dict: dict, private_key: bytes = None, env: str = None) -> T:
    config_args = dict_to_args(config_dict=config_dict, env=env, private_key=private_key)

    # Load data into dataclass
    try:
        config = config_class(**config_args)
    except Exception as e:
        raise ValueError(f"Configuration data does not match config class: {e}")

    return config


################### Loader Helpers ###################


def yaml_to_dict(file: str):
    if not path.exists(file):
        raise FileExistsError(f"Cannot read configuration file: {e}")

    # Load configuration
    try:
        with open(file) as config_file:
            config_string = config_file.read()
    except Exception as e:
        raise ValueError(f"Cannot read configuration file: {e}")

    # Parse YAML
    try:
        config_dict = yaml.load(config_string, Loader=yaml.FullLoader)
    except Exception as e:
        raise SyntaxError(f"Cannot parse configuration file: {e}")

    return config_dict


def dict_to_args(config_dict: dict, private_key: bytes = None, env: str = None):
    if not "config" in config_dict:
        raise KeyError("Config is missing config section")

    # Copy new config
    config_args = dict(config_dict["config"])

    # Layer environment data
    if env is not None:
        overlay_config_env(config_dict, config_args, env)

    # Decrypt Secrets
    if config_has_secrets(config_args):
        if not private_key:
            raise KeyError("Config has encrypted secrets, but no private key was provided")
        try:
            rsa_private_key = rsa.PrivateKey.load_pkcs1(private_key)
        except Exception as e:
            raise KeyError(f"Private key provided was invalid: {e}")

        decrypt_config_keys(config_args, rsa_private_key)

    return config_args


################### Config Helpers ###################


def config_has_secrets(config_args: dict):
    for key in config_args.keys():
        if key.endswith(".encrypted"):
            return True

    return False


def decrypt_config_keys(config_args: dict, private_key: rsa.PrivateKey):
    for key in list(config_args.keys()):
        if key.endswith(".encrypted"):
            decrypted_key = key[:-10]
            # Base64 decode
            try:
                config_args[decrypted_key] = decrypt(config_args[key], private_key)
            except Exception as e:
                raise ValueError(f"Key Error '{key}': {e}")

            del config_args[key]


def overlay_config_env(config_dict: dict, config_args: dict, env: str):
    for key in list(config_args.keys()):
        if not "env" in config_dict:
            raise KeyError("Config is missing env section")
        if not env in config_dict["env"]:
            raise KeyError(f"Config is missing {env} in env section")

        overlayed_config = config_dict["env"][env]
        for key in overlayed_config.keys():
            config_args[key] = overlayed_config[key]


################### Value Encryption/Decryption ###################


def encrypt(value: Any, public_key: rsa.PublicKey):
    return b64encode(rsa.encrypt(json.dumps(value).encode(), public_key)).decode()


def decrypt(value: str, private_key: rsa.PrivateKey):
    # Base64 decode
    try:
        decoded_value = b64decode(value)
    except Exception as e:
        raise ValueError(f"Could not decode base64 encrypted: {e}")

    # RSA Decrypt
    try:
        decrypted_value = rsa.decrypt(decoded_value, private_key).decode()
    except Exception as e:
        raise ValueError(f"Could not decrypt with provided key: {e}")

    # JSON Load
    try:
        result = json.loads(decrypted_value)
    except Exception as e:
        raise ValueError(f"Could not load JSON data: {e}")

    return result


################### Helpers ###################


def env_base64_value(key: str):
    """Helper to decode base64 environment variables such as private keys"""
    if not key in environ:
        raise KeyError(f"Missing key in environment: {key}")
    try:
        key_decoded = b64decode(environ.get(key))
    except Exception as e:
        raise ValueError(f"Could not decode environment variable {key}: {e}")

    return key_decoded



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "Click>=7.0",
    "PyYAML>=5.3.1",
    "rsa>=4.9",
]

test_requirements = [
    "pytest>=3",
]

setup(
    author="Channel Cat",
    author_email="channelcat@gmail.com",
    python_requires=">=3.9",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    description="Configuration and Secrets for Python applications",
    entry_points={
        "console_scripts": [
            "configuretron=configuretron.cli:main",
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="configuretron",
    name="configuretron",
    packages=find_packages(include=["configuretron", "configuretron.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/channelcat/configuretron",
    version="0.1.0",
    zip_safe=False,
)
